"""
Implementation of DDPG algorithm
"""
import tensorflow as tf
import numpy as np
import tflearn
#import argparse
#import pprint as pp

# Actor DNN
class ActorNetwork(object):
    #Input = s
    #Output = a

    def __init__(self, sess, stateDim, actionDim, actionBound, actionMean, learningRate, tau, minibatchSize):
        self.sess = sess
        self.sDim = stateDim
        self.aDim = actionDim
        self.aBound = actionBound
        self.aMean = actionMean
        self.learningRate = learningRate
        self.tau = tau
        self.minibatchSize = minibatchSize

        #Actor Network
        self.inputs, self.output, self.shiftedOut = self.createActorNetwork()

        self.networkParams = tf.trainable_variables()

        #Target Network
        self.targetInputs, self.targetOut, self.targetShiftedOut = self.createActorNetwork()

        self.targetNetworkParams = tf.trainable_variables()[len(self.networkParams):]

        #Periodicall updating of target network with online network weights
        self.updateTargetNetworkParams = \
            [self.targetNetworkParams[i].assign(tf.multiply(self.networkParams[i], self.tau) + 
                                                tf.multiply(self.targetNetworkParams[i], 1. - self.tau))
                for i in range(len(self.targetNetworkParams))]

        #Action gradient will be provided by critic network
        self.actionGradient = tf.placeholder(tf.float32, [None, self.aDim])

        #Combine gradients
        self.unnormalizedActorGradients = tf.gradients(
            self.shiftedOut, self.networkParams, -self.actionGradient)
        self.actorGradients = list(map(lambda x: tf.divide(x, self.minibatchSize), self.unnormalizedActorGradients))

        #Optimization
        self.optimize = tf.train.AdamOptimizer(self.learningRate).\
            apply_gradients(zip(self.actorGradients, self.networkParams))

        self.numTrainableVars = len(self.networkParams) + len(self.targetNetworkParams)


    def createActorNetwork(self):
        inputs = tflearn.input_data(shape=[None, self.sDim])
        net = tflearn.fully_connected(inputs, 400)
        net = tflearn.layers.normalization.batch_normalization(net)
        net = tflearn.activations.relu(net)
        net = tflearn.fully_connected(net, 300)
        net = tflearn.layers.normalization.batch_normalization(net)
        net = tflearn.activations.relu(net)
        w_init = tflearn.initializations.uniform(minval=-0.003, maxval=0.003)
        out = tflearn.fully_connected(net, self.aDim, activation='tanh', weights_init=w_init)
        scaledOut = tf.multiply(out, self.aBound)
        shiftedOut = tf.add(scaledOut, self.aMean)
        return inputs, out, shiftedOut

    def train(self, inputs, aGradient):
        self.sess.run(self.optimize, feed_dict={
            self.inputs: inputs,
            self.actionGradient: aGradient
        })

    def predict(self, inputs):
        return self.sess.run(self.shiftedOut, feed_dict={
            self.inputs: inputs
        })

    def predictTarget(self, inputs):
        return self.sess.run(self.targetShiftedOut, feed_dict={
            self.targetInputs: inputs
        })

    def updateTargetNetwork(self):
        self.sess.run(self.updateTargetNetworkParams)

    def getNumTrainableVars(self):
        return self.numTrainableVars

#Critic DNN
class CriticNetwork(object):
    #Input = (s,a)
    #Output = Q(s,a)

    def __init__(self, sess, stateDim, actionDim, learningRate, tau, gamma, numActorVars):
        self.sess = sess
        self.sDim = stateDim
        self.aDim = actionDim
        self.learningRate = learningRate
        self.tau = tau
        self.gamma = gamma

        #Create critic network
        self.inputs, self.action, self.out = self.createCriticNetwork()

        self.networkParams = tf.trainable_variables()[numActorVars:]

        #Target network
        self.targetInputs, self.targetAction, self.targetOut = self.createTargetCriticNetwork()

        self.targetNetworkParams = tf.trainable_variables()[(len(self.networkParams) + numActorVars):]

        #Periodically updating of target network with online network weights with regularization
        self.updateTargetNetworkParams = \
            [self.targetNetworkParams[i].assign(tf.multiply(self.networkParams[i], self.tau) \
            + tf.multiply(self.targetNetworkParams[i], 1. -self.tau))
                for i in range(len(self.targetNetworkParams))]

        #Network target (y_i)
        self.targetQvalue = tf.placeholder(tf.float32, [None, 1])

        #Define loss and optimization Op
        self.loss = tflearn.mean_square(self.targetQvalue, self.out)
        self.optimize = tf.train.AdamOptimizer(self.learningRate).minimize(self.loss)

        #Get action gradient
        self.actionGrads = tf.gradients(self.out, self.action)

    def createCriticNetwork(self):
        inputs = tflearn.input_data(shape=[None, self.sDim],name='StateInputData')
        action = tflearn.input_data(shape=[None, self.aDim],name='ActionInputData')
        net = tflearn.fully_connected(inputs, 400)
        net = tflearn.layers.normalization.batch_normalization(net)
        net = tflearn.activations.relu(net)

        #Add action tensor in 2nd layer
        t1 = tflearn.fully_connected(net, 300)
        t2 = tflearn.fully_connected(action, 300)

        net = tflearn.activation(tf.matmul(net, t1.W) + tf.matmul(action, t2.W) + t2.b, activation='relu')

        wInit = tflearn.initializations.uniform(minval=-0.003, maxval=0.003)
        out = tflearn.fully_connected(net, 1, weights_init=wInit)
        return inputs, action, out

    def createTargetCriticNetwork(self):
        inputs = tflearn.input_data(shape=[None, self.sDim],name='TargetStateInput')
        action = tflearn.input_data(shape=[None, self.aDim],name='TargetActionInput')
        net = tflearn.fully_connected(inputs, 400)
        net = tflearn.layers.normalization.batch_normalization(net)
        net = tflearn.activations.relu(net)

        #Add action tensor in 2nd layer
        t1 = tflearn.fully_connected(net, 300)
        t2 = tflearn.fully_connected(action, 300)

        net = tflearn.activation(tf.matmul(net, t1.W) + tf.matmul(action, t2.W) + t2.b, activation='relu')

        wInit = tflearn.initializations.uniform(minval=-0.003, maxval=0.003)
        out = tflearn.fully_connected(net, 1, weights_init=wInit)
        return inputs, action, out

    def train(self, inputs, action, targetQvalue):
        self.sess.run([self.out, self.optimize],feed_dict={
            self.inputs: inputs,
            self.action: action,
            self.targetQvalue: targetQvalue
        })

    def predict(self, inputs, action):
        return self.sess.run(self.out, feed_dict={
            self.inputs: inputs,
            self.action: action
            })
        
    def predictTarget(self, inputs, action):
        target = self.sess.run(self.targetOut, feed_dict={
            self.targetInputs: inputs,
            self.targetAction: action
            })
        return target

    def actionGradients(self, inputs, actions):
        return self.sess.run(self.actionGrads, feed_dict={
            self.inputs: inputs,
            self.action: actions
        })

    def updateTargetNetwork(self):
        self.sess.run(self.updateTargetNetworkParams)
